@@ -27,13 +27,15 @@ Vinay Sajip to make use of the subprocess module (Steve's version uses os.fork()
 and so does not work on Windows). Renamed to gnupg.py to avoid confusion with
 the previous versions.
 
-Modifications Copyright (C) 2008-2023 Vinay Sajip. All rights reserved.
+Modifications Copyright (C) 2008-2024 Vinay Sajip. All rights reserved.
 
 For the full documentation, see https://docs.red-dove.com/python-gnupg/ or
 https://gnupg.readthedocs.io/
 """
 
 import codecs
+from datetime import datetime
+from email.utils import parseaddr
 from io import StringIO
 import logging
 import os
@@ -43,9 +45,9 @@ from subprocess import Popen, PIPE
 import sys
 import threading
 
-__version__ = '0.5.2'
+__version__ = '0.5.4'
 __author__ = 'Vinay Sajip'
-__date__ = '$12-Dec-2023 07:52:07$'
+__date__ = '$07-Jan-2025 10:13:17$'
 
 STARTUPINFO = None
 if os.name == 'nt':  # pragma: no cover
@@ -215,6 +217,8 @@ class StatusHandler(object):
     The base class for handling status messages from `gpg`.
     """
 
+    on_data_failure = None  # set at instance level when failures occur
+
     def __init__(self, gpg):
         """
         Initialize an instance.
@@ -1000,6 +1004,55 @@ class Sign(StatusHandler, TextHandler):
             logger.debug('message ignored: %s, %s', key, value)
 
 
+class AutoLocateKey(StatusHandler):
+    """
+    This class handles status messages during key auto-locating.
+    fingerprint: str
+    key_length: int
+    created_at: date
+    email: str
+    email_real_name: str
+    """
+
+    def __init__(self, gpg):
+        StatusHandler.__init__(self, gpg)
+        self.fingerprint = None
+        self.type = None
+        self.created_at = None
+        self.email = None
+        self.email_real_name = None
+
+    def handle_status(self, key, value):
+        if key == "IMPORTED":
+            _, email, display_name = value.split()
+
+            self.email = email
+            self.email_real_name = display_name[1:-1]
+        elif key == "KEY_CONSIDERED":
+            self.fingerprint = value.strip().split()[0]
+
+    def pub(self, args):
+        """
+        Internal method to handle the 'pub' status message.
+        `pub` message contains the fingerprint of the public key, its type and its creation date.
+        """
+        pass
+
+    def uid(self, args):
+        self.created_at = datetime.fromtimestamp(int(args[5]))
+        raw_email_content = args[9]
+        email, real_name = parseaddr(raw_email_content)
+        self.email = email
+        self.email_real_name = real_name
+
+    def sub(self, args):
+        self.key_length = int(args[2])
+
+    def fpr(self, args):
+        # Only store the first fingerprint
+        self.fingerprint = self.fingerprint or args[9]
+
+
 VERSION_RE = re.compile(r'^cfg:version:(\d+(\.\d+)*)'.encode('ascii'))
 HEX_DIGITS_RE = re.compile(r'[0-9a-f]+$', re.I)
 PUBLIC_KEY_RE = re.compile(r'gpg: public key is (\w+)')
@@ -1029,6 +1082,7 @@ class GPG(object):
         'trust': TrustResult,
         'verify': Verify,
         'export': ExportResult,
+        'auto-locate-key': AutoLocateKey,
     }
     "A map of GPG operations to result object types."
 
@@ -1197,17 +1251,27 @@ class GPG(object):
         # Read the contents of the file from GPG's stdout
         assert buffer_size > 0
         chunks = []
+        on_data_failure = None
         while True:
             data = stream.read(buffer_size)
             if len(data) == 0:
                 if on_data:
-                    on_data(data)
+                    try:
+                        on_data(data)
+                    except Exception as e:
+                        if on_data_failure is None:
+                            on_data_failure = e
                 break
             if log_everything:
                 logger.debug('chunk: %r' % data[:256])
             append = True
             if on_data:
-                append = on_data(data) is not False
+                try:
+                    on_data_result = on_data(data)
+                    append = on_data_result is not False
+                except Exception as e:
+                    if on_data_failure is None:
+                        on_data_failure = e
             if append:
                 chunks.append(data)
         if _py3k:
@@ -1215,6 +1279,8 @@ class GPG(object):
             result.data = type(data)().join(chunks)
         else:
             result.data = ''.join(chunks)
+        if on_data_failure:
+            result.on_data_failure = on_data_failure
 
     def _collect_output(self, process, result, writer=None, stdin=None):
         """
@@ -1284,13 +1350,15 @@ class GPG(object):
                 stdin = codecs.getwriter(self.encoding)(p.stdin)
             else:
                 stdin = p.stdin
+            writer = None  # See issue #237
             if passphrase:
                 _write_passphrase(stdin, passphrase, self.encoding)
             writer = _threaded_copy_data(fileobj, stdin, self.buffer_size)
             self._collect_output(p, result, writer, stdin)
             return result
         finally:
-            writer.join(0.01)
+            if writer:
+                writer.join(0.01)
             if fileobj is not fileobj_or_path:
                 fileobj.close()
 
@@ -1869,6 +1937,30 @@ class GPG(object):
                 getattr(result, keyword)(L)
         return result
 
+    def auto_locate_key(self, email, mechanisms=None, **kwargs):
+        """
+        Auto locate a public key by `email`.
+
+        Args:
+            email (str): The email address to search for.
+            mechanisms (list[str]): A list of mechanisms to use. Valid mechanisms can be found
+            here https://www.gnupg.org/documentation/manuals/gnupg/GPG-Configuration-Options.html
+            under "--auto-key-locate". Default: ['wkd', 'ntds', 'ldap', 'cert', 'dane', 'local']
+        """
+        mechanisms = mechanisms or ['wkd', 'ntds', 'ldap', 'cert', 'dane', 'local']
+
+        args = ['--auto-key-locate', ','.join(mechanisms), '--locate-keys', email]
+
+        result = self.result_map['auto-locate-key'](self)
+
+        if 'extra_args' in kwargs:
+            args.extend(kwargs['extra_args'])
+
+        process = self._open_subprocess(args)
+        self._collect_output(process, result, stdin=process.stdin)
+        self._decode_result(result)
+        return result
+
     def gen_key(self, input):
         """
         Generate a key; you might use `gen_key_input()` to create the input.
@@ -2119,7 +2211,7 @@ class GPG(object):
             args.extend(extra_args)
         result = self.result_map['crypt'](self)
         self._handle_io(args, fileobj_or_path, result, passphrase, binary=True)
-        logger.debug('decrypt result[:100]: %r', result.data[:100])
+        # logger.debug('decrypt result[:100]: %r', result.data[:100])
         return result
 
     def get_recipients(self, message, **kwargs):
